﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using OpenCvSharp;
using System.Threading;
using Phidget22;
using System.IO;
using ColorMine;
using ColorMine.ColorSpaces;
using ColorMine.ColorSpaces.Comparisons;
using Newtonsoft.Json;

namespace PerlerSorter {
	public partial class Form1 : Form {

		/* Phidget object */
		Stepper top = new Stepper();
		Stepper bottom = new Stepper();
		DigitalOutput led = new DigitalOutput();
		VoltageRatioInput magSensor = new VoltageRatioInput();

		/* Motor parameters */
		const double topVelocityLimit = 2000;
		const double topAccel = 10000;
		const double topCurrentLimit = 1.4;

		const double bottomVelocityLimit = 6000;
		const double bottomAccel = 10000;
		const double bottomCurrentLimit = 1.2;

		const int stepsPerRev = 3200;

		/* Constants */
		const int beadCompartments = 19;
		/* Daylight - Source: http://www.easyrgb.com/en/math.php */
		readonly double[] xyzColorRef = {
			95.047,
			100.0,
			108.883
		};

		#region Startup, Shutdown
		public Form1() {

			InitializeComponent();

			/* Allow up to 4 listed cameras */
			for (int i = 0; i < 4; i++)
				cameraCmb.Items.Add(i);
			cameraCmb.SelectedIndex = 0;

			/* Create a colour panel for each bead compartement */
			for (int i = 0; i < beadCompartments; i++) {

				FlowLayoutPanel flp = new FlowLayoutPanel();
				flp.FlowDirection = FlowDirection.LeftToRight;
				flp.AutoSize = true;
				flp.AutoSizeMode = AutoSizeMode.GrowAndShrink;

				Button b = new Button();
				b.Text = i.ToString();
				b.Tag = i;
				b.Click += B_Click;

				Panel p = new Panel();
				p.BorderStyle = System.Windows.Forms.BorderStyle.Fixed3D;
				p.Size = new System.Drawing.Size(30, 30);
				colorPanels.Add(p);

				TextBox t = new TextBox();
				t.ReadOnly = true;
				t.Size = new System.Drawing.Size(55, 20);
				t.TextAlign = System.Windows.Forms.HorizontalAlignment.Right;
				t.Text = "0";
				colorsTxts.Add(t);

				colorCnts.Add(0);

				flp.Controls.Add(b);
				flp.Controls.Add(p);
				flp.Controls.Add(t);

				positionsPanel.Controls.Add(flp);
			}
			
			/* Read is saved colours */
			readInColors();
		}

		private void Form1_Load(object sender, EventArgs e) {

			/* Initialize and open Phidgets */
			top.HubPort = 0;
			top.Attach += Top_Attach;
			top.Detach += Top_Detach;
			top.PositionChange += Top_PositionChange;
			top.Open();

			bottom.HubPort = 1;
			bottom.Attach += Bottom_Attach;
			bottom.Detach += Bottom_Detach;
			bottom.PositionChange += Bottom_PositionChange;
			bottom.Open();

			magSensor.HubPort = 2;
			magSensor.IsHubPortDevice = true;
			magSensor.Attach += MagSensor_Attach;
			magSensor.Detach += MagSensor_Detach;
			magSensor.SensorChange += MagSensor_SensorChange;
			magSensor.Open();

			led.HubPort = 5;
			led.IsHubPortDevice = true;
			led.Channel = 0;
			led.Attach += Led_Attach;
			led.Detach += Led_Detach;
			led.Open();
		}

		private void Form1_FormClosing(object sender, FormClosingEventArgs e) {
			SetBottomPosition(0);
			Thread.Sleep(1000);
		}

		#endregion

		#region Phidget Event Handlers
		private void Bottom_PositionChange(object sender, Phidget22.Events.StepperPositionChangeEventArgs e) {
			bottomPosnTxt.Text = e.Position.ToString();
			bottomSpotTxt.Text = BottomPosition.ToString();
		}

		private void MagSensor_SensorChange(object sender, Phidget22.Events.VoltageRatioInputSensorChangeEventArgs e) {
			magSensorTxt.Text = e.SensorValue.ToString();
		}

		private void Top_PositionChange(object sender, Phidget22.Events.StepperPositionChangeEventArgs e) {
			topPosnTxt.Text = e.Position.ToString();
		}

		private void Led_Detach(object sender, Phidget22.Events.DetachEventArgs e) {
			ledAttachedChk.Checked = false;
		}

		private void Led_Attach(object sender, Phidget22.Events.AttachEventArgs e) {
			ledAttachedChk.Checked = true;
			led.State = true;
			ledChk.Checked = true;
		}

		private void MagSensor_Detach(object sender, Phidget22.Events.DetachEventArgs e) {
			magSensorAttachedChk.Checked = false;
		}

		private void MagSensor_Attach(object sender, Phidget22.Events.AttachEventArgs e) {
			magSensorAttachedChk.Checked = true;
			magSensor.SensorType = VoltageRatioSensorType.PN_1108;
			magSensor.DataInterval = 16;
		}

		private void Bottom_Detach(object sender, Phidget22.Events.DetachEventArgs e) {
			bottomAttachedChk.Checked = false;
		}

		private void Bottom_Attach(object sender, Phidget22.Events.AttachEventArgs e) {
			bottomAttachedChk.Checked = true;
			bottom.CurrentLimit = bottomCurrentLimit;
			bottom.Engaged = true;
			bottom.VelocityLimit = bottomVelocityLimit;
			bottom.Acceleration = bottomAccel;
			bottom.DataInterval = 100;
		}

		private void Top_Detach(object sender, Phidget22.Events.DetachEventArgs e) {
			topAttachedChk.Checked = false;
		}

		private void Top_Attach(object sender, Phidget22.Events.AttachEventArgs e) {
			topAttachedChk.Checked = true;
			top.CurrentLimit = topCurrentLimit;
			top.Engaged = true;
			top.RescaleFactor = -1;
			top.VelocityLimit = -topVelocityLimit;
			top.Acceleration = -topAccel;
			top.DataInterval = 100;
		}
		#endregion

		#region Motors, Sensors

		Thread alignMotorThread;
		Boolean sawMagnet;
		double magSensorMax = 0;
		private void alignMotor() {

			//Find the magnet
			top.DataInterval = top.MinDataInterval;

			sawMagnet = false;
			magSensor.SensorChange += magSensorStopMotor;
			top.VelocityLimit = -1000;

			int tryCount = 0;

			tryagain:

			top.TargetPosition += stepsPerRev;
			while (top.IsMoving && !sawMagnet)
				Thread.Sleep(25);

			if (!sawMagnet) {
				if (tryCount > 3) {
					Console.WriteLine("Align failed");
					top.Engaged = false;
					bottom.Engaged = false;
					runtest = false;
					return;
				}

				tryCount++;
				Console.WriteLine("Are we stuck? Trying a backup...");
				top.TargetPosition -= 600;
				while (top.IsMoving)
					Thread.Sleep(100);

				goto tryagain;

			}

			top.VelocityLimit = -100;
			magData = new List<KeyValuePair<double, double>>();
			magSensor.SensorChange += magSensorCollectPositionData;
			top.TargetPosition += 300;
			while (top.IsMoving)
				Thread.Sleep(100);

			magSensor.SensorChange -= magSensorCollectPositionData;

			top.VelocityLimit = -topVelocityLimit;

			KeyValuePair<double, double> max = magData[0];
			foreach (KeyValuePair<double, double> pair in magData)
				if (pair.Value > max.Value)
					max = pair;

			top.AddPositionOffset(-max.Key);
			magSensorMax = max.Value;

			top.TargetPosition = 0;
			while (top.IsMoving)
				Thread.Sleep(100);

			Console.WriteLine("Align succeeded");
		}

		List<KeyValuePair<double, double>> magData;
		private void magSensorCollectPositionData(object sender, Phidget22.Events.VoltageRatioInputSensorChangeEventArgs e) {
			magData.Add(new KeyValuePair<double, double>(top.Position, e.SensorValue));
		}

		private void magSensorStopMotor(object sender, Phidget22.Events.VoltageRatioInputSensorChangeEventArgs e) {
			if (top.IsMoving && e.SensorValue > 5) {
				top.TargetPosition = top.Position - 300;
				magSensor.SensorChange -= magSensorStopMotor;
				sawMagnet = true;
			}
		}

		private void nextMagnet(Boolean wait = false) {

			double posn = top.Position % stepsPerRev;
			top.TargetPosition += (stepsPerRev - posn);

			if (wait)
				while (top.IsMoving)
					Thread.Sleep(50);
		}

		private void nextCamera(Boolean wait = false) {
			double posn = top.Position % stepsPerRev;
			if (posn < Properties.Settings.Default.cameraOffset)
				top.TargetPosition += (Properties.Settings.Default.cameraOffset - posn);
			else
				top.TargetPosition += ((Properties.Settings.Default.cameraOffset - posn) + stepsPerRev);

			if (wait)
				while (top.IsMoving)
					Thread.Sleep(50);
		}

		private void nextHole(Boolean wait = false) {
			double posn = top.Position % stepsPerRev;
			if (posn < Properties.Settings.Default.holeOffset)
				top.TargetPosition += (Properties.Settings.Default.holeOffset - posn);
			else
				top.TargetPosition += ((Properties.Settings.Default.holeOffset - posn) + stepsPerRev);

			if (wait)
				while (top.IsMoving)
					Thread.Sleep(50);
		}

		private int BottomPosition {
			get {
				int posn = (int)bottom.Position % stepsPerRev;
				if (posn < 0)
					posn += stepsPerRev;

				return (int)Math.Round(((posn * beadCompartments) / (double)stepsPerRev));
			}
		}

		private void SetBottomPosition(int posn, bool wait = false) {
			posn = posn % beadCompartments;
			double targetPosn = (posn * stepsPerRev) / beadCompartments;

			double currentPosn = bottom.Position % stepsPerRev;
			double posnDiff = targetPosn - currentPosn;

			// Keep it as full steps
			posnDiff = ((int)(posnDiff / 16)) * 16;

			if (posnDiff <= 1600)
				bottom.TargetPosition += posnDiff;
			else
				bottom.TargetPosition -= (stepsPerRev - posnDiff);

			if (wait)
				while (bottom.IsMoving)
					Thread.Sleep(50);
		}

		private void ledChk_CheckedChanged(object sender, EventArgs e) {
			led.State = ledChk.Checked;
		}

		private void zeroBottomBtn_Click(object sender, EventArgs e) {

			bottom.AddPositionOffset(-bottom.Position);
		}

		private void alignTopBtn_Click(object sender, EventArgs e) {

			alignMotorThread = new Thread(new ThreadStart(alignMotor));
			alignMotorThread.Start();
		}

		private void topTrk_Scroll(object sender, EventArgs e) {
			top.TargetPosition = topTrk.Value;
		}

		private void bottomTrk_Scroll(object sender, EventArgs e) {
			bottom.TargetPosition = bottomTrk.Value;
		}
		private void cameraBtn_Click(object sender, EventArgs e) {
			nextCamera();
		}
		private void holeBtn_Click(object sender, EventArgs e) {
			nextHole();
		}

		private void setCameraBtn_Click(object sender, EventArgs e) {

			Properties.Settings.Default.cameraOffset = top.Position % stepsPerRev;
			Properties.Settings.Default.Save();
		}

		private void setHoleBtn_Click(object sender, EventArgs e) {

			Properties.Settings.Default.holeOffset = top.Position % stepsPerRev;
			Properties.Settings.Default.Save();
		}

		#endregion

		#region Camera

		bool runVideo = true;
		bool videoRunning = false;
		VideoCapture capture;
		Thread cvThread;
		Color detectedColor;
		Boolean detecting = false;
		int detectCnt = 0;

		private void cvThreadFunction() {

			videoRunning = false;
			capture = new VideoCapture(selectedCamera);

			using (Window window = new Window("capture")) {
				Mat image = new Mat();
				Mat image2 = new Mat();
				
				while (runVideo) {
					capture.Read(image);
					if (image.Empty())
						break;

					if (detecting)
						detectCnt++;
					else
						detectCnt = 0;

					if (detecting || circleDetectChecked || showDetectionImgChecked) {
						Cv2.CvtColor(image, image2, ColorConversionCodes.BGR2GRAY);
						Mat thres = image2.Threshold((double)Properties.Settings.Default.videoThresh, 255, ThresholdTypes.Binary);
						thres = thres.GaussianBlur(new OpenCvSharp.Size(9, 9), 10);

						if (showDetectionImgChecked)
							image = thres;

						if (detecting || circleDetectChecked) {
							CircleSegment[] bead = thres.HoughCircles(HoughMethods.Gradient, 2, /*thres.Rows/4*/ 20, 200, 100, 20, 65);
							if (bead.Length >= 1) {
								image.Circle(bead[0].Center, 3, new Scalar(0, 100, 0), -1);
								image.Circle(bead[0].Center, (int)bead[0].Radius, new Scalar(0, 0, 255), 3);
								if (bead[0].Radius >= 55) {
									Properties.Settings.Default.x = (decimal)bead[0].Center.X + (decimal)(bead[0].Radius / 2);
									Properties.Settings.Default.y = (decimal)bead[0].Center.Y - (decimal)(bead[0].Radius / 2);
								} else {
									Properties.Settings.Default.x = (decimal)bead[0].Center.X + (decimal)(bead[0].Radius);
									Properties.Settings.Default.y = (decimal)bead[0].Center.Y - (decimal)(bead[0].Radius);
								}
								Properties.Settings.Default.size = 15;
								Properties.Settings.Default.height = 15;
							} else {

								CircleSegment[] circles = thres.HoughCircles(HoughMethods.Gradient, 2, /*thres.Rows/4*/ 5, 200, 100, 60, 180);
								if (circles.Length > 1) {
									List<float> xs = circles.Select(c => c.Center.X).ToList();
									xs.Sort();
									List<float> ys = circles.Select(c => c.Center.Y).ToList();
									ys.Sort();

									int medianX = (int)xs[xs.Count / 2];
									int medianY = (int)ys[ys.Count / 2];

									if (medianX > image.Width - 15)
										medianX = image.Width - 15;
									if (medianY > image.Height - 15)
										medianY = image.Height - 15;

									image.Circle(medianX, medianY, 100, new Scalar(0, 0, 150), 3);

									if (detecting) {
										Properties.Settings.Default.x = medianX - 7;
										Properties.Settings.Default.y = medianY - 7;
										Properties.Settings.Default.size = 15;
										Properties.Settings.Default.height = 15;
									}
								}
							}
						}
					}

					Rect r = new Rect((int)Properties.Settings.Default.x,
						(int)Properties.Settings.Default.y,
						(int)Properties.Settings.Default.size,
						(int)Properties.Settings.Default.height);

					Mat beadSample = new Mat(image, r);
					Scalar avgColor = Cv2.Mean(beadSample);
					detectedColor = Color.FromArgb((int)avgColor[2], (int)avgColor[1], (int)avgColor[0]);

					image.Rectangle(r, new Scalar(0, 150, 0));

					window.ShowImage(image);
					Cv2.WaitKey(1);
					videoRunning = true;
				}

				videoRunning = false;
			}
		}

		int selectedCamera = -1;
		private void cameraCmb_SelectedIndexChanged(object sender, EventArgs e) {
			selectedCamera = cameraCmb.SelectedIndex;
		}

		private void cameraStartBtn_Click(object sender, EventArgs e) {

			if (cameraStartBtn.Text == "start") {
				cvThread = new Thread(new ThreadStart(cvThreadFunction));
				runVideo = true;
				cvThread.Start();
				cameraStartBtn.Text = "stop";
				while (!videoRunning)
					Thread.Sleep(100);

				updateColorTimer.Start();

			} else {
				runVideo = false;
				cvThread.Join();
				cameraStartBtn.Text = "start";
			}
		}

		private void settingsBtn_Click(object sender, EventArgs e) {
			capture.Set(CaptureProperty.Settings, 1);
		}

		bool circleDetectChecked = false;
		private void checkBox1_CheckedChanged(object sender, EventArgs e) {
			circleDetectChecked = showCirclesChk.Checked;
		}

		bool showDetectionImgChecked = false;
		private void showDetectionImgChk_CheckedChanged(object sender, EventArgs e) {
			showDetectionImgChecked = showDetectionImgChk.Checked;
		}

		private void thresholdUpDown_ValueChanged(object sender, EventArgs e) {
			Properties.Settings.Default.Save();
		}
		#endregion

		#region Colours

		List<Color> colors = new List<Color>();
		List<Panel> colorPanels = new List<Panel>();
		List<TextBox> colorsTxts = new List<TextBox>();
		List<int> colorCnts = new List<int>();

		const int numColorSpots = 18;
		const int unknownColorIndex = 18;
		int findColorPosition(Color c) {

			Console.WriteLine("Finding color...");

			var cRGB = new Rgb();
			cRGB.R = c.R;
			cRGB.G = c.G;
			cRGB.B = c.B;

			int bestMatch = -1;
			double matchDelta = 100;

			for (int i = 0; i < colors.Count; i++) {

				var RGB = new Rgb();
				RGB.R = colors[i].R;
				RGB.G = colors[i].G;
				RGB.B = colors[i].B;

				double delta = cRGB.Compare(RGB, new CieDe2000Comparison());
				//double delta = deltaE(c, colors[i]);
				Console.WriteLine("DeltaE (" + i.ToString() + "): " + delta.ToString());
				if (delta < matchDelta) {
					matchDelta = delta;
					bestMatch = i;
				}
			}

			if (matchDelta < 5) {
				Console.WriteLine("Found! (Posn: " + bestMatch + " Delta: " + matchDelta + ")");
				return bestMatch;
			}

			if (colors.Count < numColorSpots) {
				Console.WriteLine("New Color!");
				colors.Add(c);
				this.BeginInvoke(new Action<int>(setBackColor), new object[] { colors.Count - 1 });
				writeOutColors();
				return (colors.Count - 1);
			} else {
				Console.WriteLine("Unknown Color!");
				return unknownColorIndex;
			}
		}
		void setBackColor(int i) {

			colorPanels[i].BackColor = colors[i];
		}
		void writeOutColors() {
			string json = JsonConvert.SerializeObject(colors);
			using (StreamWriter sw = new StreamWriter(new FileStream("colors.json", FileMode.Create))) {
				sw.Write(json);
			}
		}

		void readInColors() {
			try {
				string json;
				FileInfo file = new FileInfo("colors.json");
				if (file.Exists) {
					using (StreamReader sr = new StreamReader("colors.json")) {
						json = sr.ReadToEnd();
					}
					colors = JsonConvert.DeserializeObject<List<Color>>(json);
				} else {
					colors = new List<Color>();
				}
			} catch {
				colors = new List<Color>();
			}

			for (int i = 0; i < colors.Count; i++) {
				colorPanels[i].BackColor = colors[i];
			}
		}

		void setColorTxt(int i) {
			colorsTxts[i].Text = colorCnts[i].ToString();
		}
		void setColorDet(Color c) {
			detColorPanel.BackColor = c;
		}

		private void resetColoursBtn_Click(object sender, EventArgs e) {
			colors = new List<Color>();
			writeOutColors();
			for (int i = 0; i < beadCompartments; i++) {
				colorPanels[i].BackColor = Color.Empty;
			}
		}

		private void resetUnknownBtn_Click(object sender, EventArgs e) {
			colorCnts[unknownColorIndex] = 0;
		}

		/* Colour position button handler */
		private void B_Click(object sender, EventArgs e) {
			Button b = (Button)sender;
			int posn = (int)b.Tag;
			SetBottomPosition(posn);
		}

		private void updateColorTimer_Tick(object sender, EventArgs e) {

			colourPanel.BackColor = detectedColor;
		}
		#endregion

		#region Main Program Logic

		Thread colourTestThread;
		Boolean runtest = false;
		void colourTest() {

			if (!top.Engaged)
				top.Engaged = true;

			if (!bottom.Engaged)
				bottom.Engaged = true;

			while (runtest) {

				nextMagnet(true);
				Thread.Sleep(100);
				try {
					if (magSensor.SensorValue < (magSensorMax - 4))
						alignMotor();
				} catch { alignMotor(); }

				nextCamera(true);

				detecting = true;
				while (detectCnt < 5)
					Thread.Sleep(25);
				Console.WriteLine("Detect Count: " + detectCnt);
				detecting = false;

				Color c = detectedColor;
				this.BeginInvoke(new Action<Color>(setColorDet), new object[] { c });
				int i = findColorPosition(c);

				SetBottomPosition(i, true);
				nextHole(true);
				colorCnts[i]++;
				this.BeginInvoke(new Action<int>(setColorTxt), new object[] { i });
				Thread.Sleep(250);

				if (colorCnts[unknownColorIndex] > 500) {
					top.Engaged = false;
					bottom.Engaged = false;
					runtest = false;
					this.BeginInvoke(new Action(setGoGreen), null);
					return;
				}
			}
		}

		private void colourTestBtn_Click(object sender, EventArgs e) {
			if (colourTestThread == null || !colourTestThread.IsAlive) {
				colourTestThread = new Thread(new ThreadStart(colourTest));
				runtest = true;
				colourTestThread.Start();
				colourTestBtn.Text = "STOP";
				colourTestBtn.BackColor = Color.Red;
			} else {
				runtest = false;
				colourTestBtn.Text = "GO";
				colourTestBtn.BackColor = Color.Green;
			}
		}
		void setGoGreen() {
			colourTestBtn.Text = "GO";
			colourTestBtn.BackColor = Color.Green;
		}

		#endregion
	}
}
